# coding: utf-8

# Copyright (c) 2007-2008 Forest Bond.
# This file is part of the pytagsfs software package.
#
# pytagsfs is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

import sys, os, time, traceback, stat
from unittest import TestCase, main

from sclapp.util import importName
from sclapp.shell import Shell, CommandFailed

from common import copy_private_inherited_test_methods

RACE_DELAY = 2

class ErrorOnError(Exception):
    def __init__(self, original_traceback, new_traceback):
        self.original_traceback = original_traceback
        self.new_traceback = new_traceback
        Exception.__init__(self, original_traceback, new_traceback)

    def __str__(self):
        return '\n'.join([
          '*** Caught an exception while cleaning up after an exception. ***',
          '',
          '=== Traceback for original exception ===',
          self.original_traceback,
          '',
          '=== Traceback for new exception ===',
          self.new_traceback,
        ])

    mount_parameters = None

TEST_NON_DEFAULT_COMPONENTS = True

def iter_mount_parameters():
    yield {
      'sourcetreerep': None,
      'sourcetreemon': None,
      'pathstore': None,
    }
    if TEST_NON_DEFAULT_COMPONENTS:
        sourcetreereps = (
          # This is more-or-less an abstract base class.  It mostly works, but
          # source tree updates are only supported by subclasses.
          #'pytagsfs.sourcetreerep.SourceTreeRepresentation',

          'pytagsfs.sourcetreerep.pollinline.'
            'PollInLineSourceTreeRepresentation',
        )
        sourcetreemons = (
          'pytagsfs.sourcetreemon.pyinotify_.PyInotifySourceTreeMonitor',
          'pytagsfs.sourcetreemon.pyinotify_.'
            'DeferredPyInotifySourceTreeMonitor',
          'pytagsfs.sourcetreemon.gamin_.GaminSourceTreeMonitor',
          'pytagsfs.sourcetreemon.gamin_.DeferredGaminSourceTreeMonitor',
        )
        pathstores = (
          'pytagsfs.pathstore.pytypes.PyTypesPathStore',
        )
        for sourcetreerep in sourcetreereps:
            for sourcetreemon in sourcetreemons:
                for pathstore in pathstores:
                    yield {
                      'sourcetreerep': sourcetreerep,
                      'sourcetreemon': sourcetreemon,
                      'pathstore': pathstore,
                    }

def dotted_last(s):
    return s.rpartition('.')[-1]

def define_integration_test_classes(base_class):
    mod = importName(__name__)

    for mount_parameters in iter_mount_parameters():
        name_parts = [base_class.__name__]

        if mount_parameters.get('sourcetreerep', None):
            name_parts.append(
              dotted_last(mount_parameters['sourcetreerep']))

        if mount_parameters.get('sourcetreemon', None):
            name_parts.append(
              dotted_last(mount_parameters['sourcetreemon']))

        if mount_parameters.get('pathstore', None):
            name_parts.append(
              dotted_last(mount_parameters['pathstore']))

        if len(name_parts) == 1:
            name_parts.append('defaults')

        subclass_name = '__'.join(name_parts)

        class subclass(base_class):
            def get_mount_parameters(self):
                return mount_parameters

        copy_private_inherited_test_methods(subclass)

        subclass.__name__ = subclass_name

        setattr(mod, subclass_name, subclass)

class _BaseIntegrationTestCase(TestCase):
    sh = None
    _ran_test_function = None

    python_executable = 'python'
    pytagsfs_executable = None

    def setUp(self):
        self._ran_test_function = False

        package_root = os.path.dirname(
          os.path.dirname(os.path.abspath(__file__)))
        modules_root = os.path.join(package_root, 'modules')

        self.sh = Shell()
        self.sh.env['PATH'] = '%s:%s' % (package_root, self.sh.env['PATH'])
        self.sh.env['PYTHONPATH'] = '%s:%s' % (
          modules_root, self.sh.env['PYTHONPATH'])

        self.pytagsfs_executable = os.path.join(package_root, 'pytagsfs')

    def tearDown(self):
        if not self._ran_test_function:
            raise AssertionError(
              'test function not called (use runTestFunction)')
        self._ran_test_function = False
        self.sh.exit()
        del self.sh

    def get_opts(self):
        raise NotImplementedError

    def get_mount_parameters(self):
        raise NotImplementedError

    def mount(self):
        mount_parameters = dict(self.get_mount_parameters())

        extra_options = []
        if mount_parameters.get('sourcetreerep', None):
            extra_options.append(
              'sourcetreerep=%s' % mount_parameters['sourcetreerep'])
        if mount_parameters.get('sourcetreemon', None):
            extra_options.append(
              'sourcetreemon=%s' % mount_parameters['sourcetreemon'])
        if mount_parameters.get('pathstore', None):
            extra_options.append(
              'pathstore=%s' % mount_parameters['pathstore'])
        extra_options.append(self.get_opts())

        self.sh.execute(
          ('? ? -o '
           'entry_timeout=0,negative_timeout=0,attr_timeout=0,debug,%s '
           'source mnt >logfile 2>&1 &'
          ) % ','.join(extra_options),
          self.python_executable,
          self.pytagsfs_executable,
        )

    def umount(self):
        self.sh.execute('fusermount -u mnt')

    def runInTestTree(self, fn):
        test_tree_root = 'test-trees'
        mount_parameters = self.get_mount_parameters()
        test_tree_subdir = '%s.%s' % (
          self.__class__.__name__,
          self._testMethodName,
        )
        test_tree = os.path.join(test_tree_root, test_tree_subdir)
        self.sh.execute('mkdir -p ?', test_tree)

        try:
            wd = os.getcwd()
            os.chdir(test_tree)
            self.sh.pushd(test_tree)
            try:
                fn()
            finally:
                self.sh.popd()
            self.sh.execute('rmdir ?', test_tree)
        finally:
            os.chdir(wd)
        self.sh.execute('rmdir ? || true', test_tree_root)

    def buildTree(self):
        self.sh.execute('mkdir -p source/a/b source/a source/x')
        self.sh.execute(u'touch source/a/b/c source/a/• source/x/y')
        self.sh.execute(u'chmod 0644 source/a/b/c source/a/• source/x/y')

    def runTestFunction(self, fn):
        self._ran_test_function = True

        def fn2():
            self.sh.execute('mkdir -p mnt source')
            self.buildTree()
            self.mount()
            time.sleep(RACE_DELAY)
            try:
                fn()
            except:
                original_traceback = traceback.format_exc()
                try:
                    self.recordFailure()
                except Exception:
                    print >>sys.stderr, traceback.format_exc()
                try:
                    self.umount()
                except Exception:
                    new_traceback = traceback.format_exc()
                    raise ErrorOnError(
                      original_traceback,
                      new_traceback,
                    )
                raise
            self.umount()
            self.sh.execute('rm -f logfile')
            self.sh.execute('rm -R mnt source')

        self.runInTestTree(fn2)

    def recordFailure(self):
        try:
            self.sh.execute(
              '( find -print0 | xargs -0 -n1 ls -lhd ) >listing 2>&1')
        except CommandFailed:
            print >>sys.stderr, traceback.format_exc()

    def assertFileExists(self, filename):
        try:
            f = open(filename, 'r')
        except (OSError, IOError), e:
            raise AssertionError(str(e))
        else:
            f.close()

    def assertFileDoesNotExist(self, filename):
        def test_fn():
            f = open(filename, 'r')
            f.close()
        self.assertRaises(IOError, test_fn)

    def assertFileContents(self, filename, expected):
        f = open(filename, 'r')
        try:
            actual = f.read().decode('utf-8', 'replace')
        finally:
            f.close()
        assert (actual == expected), (
          u'Expected:\n%s\nGot:\n%s' % (expected, actual)
        )

    def assertFileContains(self, filename, s):
        f = open(filename, 'r')
        try:
            contents = f.read().decode('utf-8', 'replace')
        finally:
            f.close()
        assert (s in contents), (
          u'Expected to find:\n%s\nGot:\n%s' % (s, contents)
        )

    def _assertFilePredicate(self, filename1, filename2, predicate):
        f1 = open(filename1, 'r')
        try:
            f2 = open(filename2, 'r')
            try:
                assert predicate(f1, f2)
            finally:
                f2.close()
        finally:
            f1.close()

    def assertFilesAreSame(self, filename1, filename2):
        self._assertFilePredicate(
          filename1, filename2, (lambda f1, f2: f1.read() == f2.read()))

    def assertFilesAreDifferent(self, filename1, filename2):
        self._assertFilePredicate(
          filename1, filename2, (lambda f1, f2: f1.read() != f2.read()))

    def assertDirectoryExists(self, dirname):
        try:
            os.listdir(dirname)
        except (IOError, OSError):
            raise AssertionError(unicode(e))

    def assertDirectoryDoesNotExist(self, dirname):
        def test_fn():
            os.listdir(dirname)
        self.assertRaises((IOError, OSError), test_fn)

class SimpleFilenamePatternTestCase(_BaseIntegrationTestCase):
    def get_opts(self):
        return 'format="/%f"'

    def _testReadDir(self):
        def fn():
            self.assertFileExists(u'mnt/c')
            self.assertFileExists(u'mnt/y')
            self.assertFileExists(u'mnt/•')
        self.runTestFunction(fn)

    def _testAddEmptyFile(self):
        def fn():
            self.sh.execute('touch source/a/b/d')
            time.sleep(RACE_DELAY)
            self.assertFileExists('mnt/d')
        self.runTestFunction(fn)

define_integration_test_classes(SimpleFilenamePatternTestCase)

class SourceExcludeTestCase(_BaseIntegrationTestCase):
    def get_opts(self):
        return 'format="/%f",srcfilter="!\/c$"'

    def _testSourceExclusion(self):
        def fn():
            status, output = self.sh.execute('find mnt')
            files = [l.strip() for l in output.strip().split('\n')]
            assert (u'/c' not in files), repr(files)
        self.runTestFunction(fn)

define_integration_test_classes(SourceExcludeTestCase)

class DestExcludeTestCase(_BaseIntegrationTestCase):
    def get_opts(self):
        return 'format="/%f",srcfilter="!\/c$"'

    def _testDestExclusion(self):
        def fn():
            status, output = self.sh.execute('find mnt')
            files = [l.strip() for l in output.strip().split('\n')]
            assert (u'/c' not in files), repr(files)
        self.runTestFunction(fn)

define_integration_test_classes(DestExcludeTestCase)

class UpdatesTestCase(_BaseIntegrationTestCase):
    def get_opts(self):
        return 'format="/%p/%f"'

    def _testUpdateFileContentsAtSource(self):
        def fn():
            content = 'foo\n'

            f = open('source/a/b/c', 'w')
            f.write(content)
            f.close()

            time.sleep(RACE_DELAY)

            self.assertFileContents('mnt/b/c', content)
        self.runTestFunction(fn)

    def _testWriteTruncate(self):
        def fn():
            f = open('mnt/b/c', 'w')
            f.write('foo\n')
            f.close()

            time.sleep(RACE_DELAY)

            self.assertFileContents('mnt/b/c', 'foo\n')
        self.runTestFunction(fn)

    def _testAppend(self):
        def fn():
            f = open('mnt/b/c', 'w')
            f.write('foo\nbar\nbaz\n')
            f.close()

            time.sleep(RACE_DELAY)

            self.assertFileContents('mnt/b/c', 'foo\nbar\nbaz\n')

            f = open('mnt/b/c', 'a')
            f.write('buz\n')
            f.close()

            time.sleep(RACE_DELAY)

            self.assertFileContents('mnt/b/c', 'foo\nbar\nbaz\nbuz\n')

        self.runTestFunction(fn)

    def _testOverwriteSection(self):
        def fn():
            f = open('mnt/b/c', 'w')
            f.write('foo\nbar\nbaz\n')
            f.close()

            time.sleep(RACE_DELAY)

            self.assertFileContents('mnt/b/c', 'foo\nbar\nbaz\n')

            fd = os.open('mnt/b/c', os.O_RDWR)
            os.lseek(fd, 4, 0)
            os.write(fd, 'biz\n')
            os.close(fd)

            time.sleep(RACE_DELAY)

            self.assertFileContents('mnt/b/c', 'foo\nbiz\nbaz\n')

        self.runTestFunction(fn)

    def _testFtruncate(self):
        def fn():
            initial = 'foo\nbar\nbaz\nbiz\nbang\nboom'
            final = initial[:8]

            f = open('source/a/m', 'w')
            f.write(initial)
            f.close()

            time.sleep(RACE_DELAY)
            self.assertFileContents('mnt/a/m', initial)

            f = open('mnt/a/m', 'r+')
            try:
                self.assertEqual(f.read(), initial)
                f.truncate(8)
                f.seek(0)
                self.assertEqual(f.read(), final)
            finally:
                f.close()

            time.sleep(RACE_DELAY)
            self.assertFileContents('mnt/a/m', final)

        self.runTestFunction(fn)

    def _testSeekToEnd(self):
        def fn():
            content = 'foo\nbar\nbaz\nbiz\nbang\nboom'

            f = open('source/a/m', 'w')
            try:
                f.write(content)
            finally:
                f.close()

            time.sleep(RACE_DELAY)
            self.assertFileContents('mnt/a/m', content)

            f = open('mnt/a/m', 'rb+')
            try:
                self.assertEqual(f.read(), content)
                f.seek(0, 2)
                self.assertEqual(f.tell(), len(content))
            finally:
                f.close()

            self.assertFileContents('mnt/a/m', content)
        self.runTestFunction(fn)

    def _testWritePastEndAndSeekToEnd(self):
        def fn():
            content = 'foo\nbar\n'
            last = 'baz\n'
            final = content + last

            f = open('source/a/m', 'w')
            try:
                f.write(content)
            finally:
                f.close()

            time.sleep(RACE_DELAY)
            self.assertFileContents('mnt/a/m', content)

            f = open('mnt/a/m', 'rb+')
            try:
                self.assertEqual(f.read(), content)
                f.seek(0, 2)
                f.write(last)
                f.seek(0)
                self.assertEqual(f.read(), final)
                f.seek(0, 2)
                self.assertEqual(f.tell(), len(final))
            finally:
                f.close()

            time.sleep(RACE_DELAY)
            self.assertFileContents('mnt/a/m', final)
        self.runTestFunction(fn)

    def _makeUpdateMtimeTestFunction(self, update_at_source):
        def fn():
            stat_a_src = os.stat('source/a/b/c')
            stat_a_dst = os.stat('mnt/b/c')

            stat_a_dst_dir = os.stat('mnt/b')

            self.assertEqual(stat_a_src.st_mtime, stat_a_dst.st_mtime)
            self.assertEqual(stat_a_src.st_mtime, stat_a_dst_dir.st_mtime)

            # Make sure enough time passes for timestamps to change.
            time.sleep(1)

            if update_at_source:
                os.utime('source/a/b/c', None)
            else:
                os.utime('mnt/b/c', None)

            time.sleep(RACE_DELAY)

            stat_b_src = os.stat('source/a/b/c')
            stat_b_dst = os.stat('mnt/b/c')

            stat_b_dst_dir = os.stat('mnt/b')

            self.assertEqual(stat_b_src.st_mtime, stat_b_dst.st_mtime)
            self.assertEqual(stat_b_src.st_mtime, stat_b_dst_dir.st_mtime)

            assert stat_a_src.st_mtime < stat_b_src.st_mtime
            assert stat_a_dst.st_mtime < stat_b_dst.st_mtime

        return fn

    def _testUpdateMtimeAtSource(self):
        fn = self._makeUpdateMtimeTestFunction(True)
        self.runTestFunction(fn)

    def _testUpdateMtimeAtDest(self):
        fn = self._makeUpdateMtimeTestFunction(False)
        self.runTestFunction(fn)

    def _makeUpdateModeTestFunction(self, update_at_source):
        def fn():
            stat_a_src = os.stat('source/a/b/c')
            stat_a_dst = os.stat('mnt/b/c')

            mode_a_src = stat.S_IMODE(stat_a_src.st_mode)
            mode_a_dst = stat.S_IMODE(stat_a_dst.st_mode)

            self.assertEqual(mode_a_src, 0644)
            self.assertEqual(mode_a_dst, 0644)

            new_mode = (mode_a_src & 0000) | 0600

            if update_at_source:
                os.chmod('source/a/b/c', new_mode)
            else:
                os.chmod('mnt/b/c', new_mode)

            time.sleep(RACE_DELAY)

            stat_b_src = os.stat('source/a/b/c')
            stat_b_dst = os.stat('mnt/b/c')

            mode_b_src = stat.S_IMODE(stat_b_src.st_mode)
            mode_b_dst = stat.S_IMODE(stat_b_dst.st_mode)

            self.assertEqual(mode_b_src, 0600)
            self.assertEqual(mode_b_dst, 0600)

        return fn

    def _testUpdateModeAtSource(self):
        fn = self._makeUpdateModeTestFunction(True)
        self.runTestFunction(fn)

    def _testUpdateModeAtDest(self):
        fn = self._makeUpdateModeTestFunction(False)
        self.runTestFunction(fn)

define_integration_test_classes(UpdatesTestCase)

################################################################################

class _BaseTagsTestCase(_BaseIntegrationTestCase):
    ext = None

    artist_name = 'foo'
    artist_name_alt = 'baz'
    track_name = 'bar'
    backup_file = 'backup'
    source_file_base = 'silence'

    artist_name_2 = 'bink'
    track_name_2 = 'bonk'

    def get_source_filename(self):
        return 'source/%s.%s' % (self.source_file_base, self.ext)

    def get_source_filename_2(self):
        return 'source/%s2.%s' % (self.source_file_base, self.ext)

    def get_source_filename_3(self):
        return 'source/%s3.%s' % (self.source_file_base, self.ext)

    def get_dest_filename(self):
        return 'mnt/%s - %s.%s' % (self.artist_name, self.track_name, self.ext)

    def get_dest_filename_alt(self):
        return 'mnt/%s - %s.%s' % (
          self.artist_name_alt, self.track_name, self.ext)

    def get_opts(self):
        return 'format="/%a - %t.%e"'

    def buildTree(self):
        data_file = '../../data/silence.%s' % self.ext
        source_file = self.get_source_filename()
        source_file_2 = self.get_source_filename_2()

        self.sh.execute('cp ? ?', data_file, source_file)
        self.sh.execute(
          'pytags --set artist=? --set title=? ?',
          self.artist_name,
          self.track_name,
          source_file,
        )

        self.sh.execute('cp ? ?', data_file, source_file_2)

    def _makeTagChangeFunction(self, change_dest):
        def fn():
            source_file = self.get_source_filename()

            dest_file_normal = self.get_dest_filename()
            dest_file_renamed = self.get_dest_filename_alt()

            self.assertFilesAreSame(source_file, dest_file_normal)

            # Make a backup so we can compare later...
            self.sh.execute('cp ? ?', dest_file_normal, self.backup_file)

            self.assertFilesAreSame(dest_file_normal, self.backup_file)

            if change_dest:
                file = dest_file_normal
            else:
                file = source_file
            self.sh.execute(
              'pytags --set artist=? ?', self.artist_name_alt, file)

            time.sleep(RACE_DELAY)

            self.assertFileExists(dest_file_renamed)
            self.assertFileDoesNotExist(dest_file_normal)

            if change_dest:
                file = dest_file_renamed
            else:
                file = source_file
            self.sh.execute('pytags --set artist=? ?', self.artist_name, file)

            time.sleep(RACE_DELAY)

            self.assertFileExists(dest_file_normal)
            self.assertFileDoesNotExist(dest_file_renamed)

            try:
                # Compare with backup...
                self.assertFilesAreSame(dest_file_normal, self.backup_file)
            except Exception:
                self.sh.execute('cp -a ? dest_file_copy', dest_file_normal)
                raise
            os.unlink(self.backup_file)
        return fn

    def _testTagChangeOnSourceFileCausingPathChange(self):
        self.runTestFunction(self._makeTagChangeFunction(False))

    def _testTagChangeOnDestFileCausingPathChange(self):
        self.runTestFunction(self._makeTagChangeFunction(True))

    def _testTagChangeOnSourceFileCausingNewFakePath(self):
        def fn():
            source_file = self.get_source_filename_2()
            dest_file = 'mnt/%s - %s.%s' % (
              self.artist_name_2, self.track_name_2, self.ext)
            self.assertFileDoesNotExist(dest_file)
            self.sh.execute(
              'pytags --set artist=? --set title=? ?',
              self.artist_name_2,
              self.track_name_2,
              source_file,
            )
            time.sleep(RACE_DELAY)
            self.assertFileExists(dest_file)
        self.runTestFunction(fn)

    def _testConflict(self):
        def fn():
            source_file = self.get_source_filename()
            source_dupe_file = self.get_source_filename_3()
            dest_file = self.get_dest_filename()

            self.assertFilesAreSame(dest_file, source_file)
            self.sh.execute('cp ? ?', source_file, source_dupe_file)
            time.sleep(RACE_DELAY)
            self.assertFilesAreSame(dest_file, source_file)
            self.assertFilesAreSame(dest_file, source_dupe_file)
            self.sh.execute('rm ?', source_file)
            time.sleep(RACE_DELAY)
            self.assertFilesAreSame(dest_file, source_dupe_file)
        self.runTestFunction(fn)

class OggTagsTestCase(_BaseTagsTestCase):
    ext = 'ogg'

define_integration_test_classes(OggTagsTestCase)

class FlacTagsTestCase(_BaseTagsTestCase):
    ext = 'flac'

define_integration_test_classes(FlacTagsTestCase)

class Mp3TagsTestCase(_BaseTagsTestCase):
    ext = 'mp3'

define_integration_test_classes(Mp3TagsTestCase)

################################################################################

class _NonAsciiSourceFileBaseTagsTestCase(_BaseTagsTestCase):
    source_file_base = u'•'

class NonAsciiSourceFileOggTagsTestCase(_NonAsciiSourceFileBaseTagsTestCase):
    ext = 'ogg'

define_integration_test_classes(NonAsciiSourceFileOggTagsTestCase)

class NonAsciiSourceFileFlacTagsTestCase(_NonAsciiSourceFileBaseTagsTestCase):
    ext = 'flac'

define_integration_test_classes(NonAsciiSourceFileFlacTagsTestCase)

class NonAsciiSourceFileMp3TagsTestCase(_NonAsciiSourceFileBaseTagsTestCase):
    ext = 'mp3'

define_integration_test_classes(NonAsciiSourceFileMp3TagsTestCase)

################################################################################

class _NonAsciiArtistBaseTagsTestCase(_BaseTagsTestCase):
    artist_name = u'•'

class NonAsciiArtistOggTagsTestCase(_NonAsciiArtistBaseTagsTestCase):
    ext = 'ogg'

define_integration_test_classes(NonAsciiArtistOggTagsTestCase)

class NonAsciiArtistFlacTagsTestCase(_NonAsciiArtistBaseTagsTestCase):
    ext = 'flac'

define_integration_test_classes(NonAsciiArtistFlacTagsTestCase)

class NonAsciiArtistMp3TagsTestCase(_NonAsciiArtistBaseTagsTestCase):
    ext = 'mp3'

define_integration_test_classes(NonAsciiArtistMp3TagsTestCase)

################################################################################

class _BaseRenameTestCase(_BaseIntegrationTestCase):
    artist_name = 'foo'
    artist_name_alt = 'baz'
    track_name = 'bar'
    track_name_alt = 'bim'
    backup_file = 'backup'
    source_file_base = 'silence'

    artist_name_2 = 'munch'
    track_name_2 = 'crunch'

    track_name_3 = 'point'

    artist_name_4 = 'jynx'

    def get_data_filename(self):
        return '../../data/silence.%s' % self.ext

    def get_source_filename(self):
        return 'source/%s.%s' % (self.source_file_base, self.ext)

    def get_source_filename_2(self):
        return 'source/%s_2.%s' % (self.source_file_base, self.ext)

    def get_source_filename_3(self):
        return 'source/%s_3.%s' % (self.source_file_base, self.ext)

    def get_dest_filename(self):
        return 'mnt/%s/%s.%s' % (self.artist_name, self.track_name, self.ext)

    def get_opts(self):
        return 'format="/%a/%t.%e"'

    def buildTree(self):
        data_file = self.get_data_filename()

        source_file = self.get_source_filename()
        self.sh.execute('cp ? ?', data_file, source_file)
        self.sh.execute(
          'pytags --set artist=? --set title=? ?',
          self.artist_name,
          self.track_name,
          source_file,
        )

        source_file_2 = self.get_source_filename_2()
        self.sh.execute('cp ? ?', data_file, source_file_2)
        self.sh.execute(
          'pytags --set artist=? --set title=? ?',
          self.artist_name_2,
          self.track_name_2,
          source_file_2,
        )

        source_file_3 = self.get_source_filename_3()
        self.sh.execute('cp ? ?', data_file, source_file_3)
        self.sh.execute(
          'pytags --set artist=? --set title=? ?',
          self.artist_name,
          self.track_name_3,
          source_file_3,
        )

    def _makeSuccessfulRenameFunction(
      self, dest_file_renamed, field, value, value_alt):
        def fn():
            source_file = self.get_source_filename()
            dest_file_normal = self.get_dest_filename()

            self.sh.execute('pytags ? | grep ?=?', source_file, field, value)
            self.sh.execute('ls ?', dest_file_normal)
            self.sh.execute( 'mv ? ?', dest_file_normal, dest_file_renamed)

            time.sleep(RACE_DELAY)

            self.sh.execute('ls ?', dest_file_renamed)
            self.sh.execute(
              'pytags ? | grep ?=?', source_file, field, value_alt)
            self.sh.execute('mv ? ?', dest_file_renamed, dest_file_normal)

            time.sleep(RACE_DELAY)

            self.sh.execute(
              'pytags ? | grep ?=?', source_file, field, value)
            self.sh.execute('ls ?', dest_file_normal)
        return fn

    def _testSimpleRename(self):
        fn = self._makeSuccessfulRenameFunction(
          'mnt/%s/%s.%s' % (self.artist_name, self.track_name_alt, self.ext),
          'title',
          self.track_name,
          self.track_name_alt,
        )
        self.runTestFunction(fn)

    def _testRenameAcrossDirectories(self):
        fn = self._makeSuccessfulRenameFunction(
          'mnt/%s/%s.%s' % (self.artist_name_2, self.track_name, self.ext),
          'artist',
          self.artist_name,
          self.artist_name_2,
        )
        self.runTestFunction(fn)

    def _makeSuccessfulDirectoryRenameFunction(
      self, dest_file_renamed, field, value, value_alt):
        def fn():
            source_file = self.get_source_filename()
            dest_file_normal = self.get_dest_filename()

            dest_dir_normal = os.path.dirname(dest_file_normal)
            dest_dir_renamed = os.path.dirname(dest_file_renamed)

            self.sh.execute('pytags ? | grep ?=?', source_file, field, value)
            self.sh.execute('ls ?', dest_file_normal)

            self.sh.execute('mv ? ?', dest_dir_normal, dest_dir_renamed)

            time.sleep(RACE_DELAY)

            self.sh.execute('ls ?', dest_file_renamed)
            self.sh.execute(
              'pytags ? | grep ?=?', source_file, field, value_alt)
            self.sh.execute('mv ? ?', dest_dir_renamed, dest_dir_normal)

            time.sleep(RACE_DELAY)

            self.sh.execute(
              'pytags ? | grep ?=?', source_file, field, value)
            self.sh.execute('ls ?', dest_file_normal)
        return fn

    def _testDirectoryRename(self):
        fn = self._makeSuccessfulDirectoryRenameFunction(
          'mnt/%s/%s.%s' % (self.artist_name_4, self.track_name, self.ext),
          'artist',
          self.artist_name,
          self.artist_name_4,
        )
        self.runTestFunction(fn)

    def _makeInvalidRenameFunction(self, dest_file_renamed):
        def fn():
            source_file = self.get_source_filename()
            dest_file_normal = self.get_dest_filename()

            # Make a backup so we can compare later...
            self.sh.execute('cp ? ?', dest_file_normal, self.backup_file)

            self.sh.execute(
              'pytags ? | grep artist=?', source_file, self.artist_name)
            self.sh.execute('ls ?', dest_file_normal)
            self.assertRaises(
              CommandFailed,
              self.sh.execute,
              'mv ? ?',
              dest_file_normal,
              dest_file_renamed,
            )

            time.sleep(RACE_DELAY)

            self.assertRaises(
              CommandFailed, self.sh.execute, 'ls ?', dest_file_renamed)
            self.assertRaises(
              CommandFailed,
              self.sh.execute,
              'pytags ? | grep artist=?',
              source_file,
              self.artist_name_alt,
            )

            try:
                # Compare with backup...
                self.assertFilesAreSame(source_file, self.backup_file)
            except Exception:
                self.sh.execute('cp -a ? source_file_copy', source_file)
                raise
            os.unlink(self.backup_file)
        return fn

    def _testInvalidDueToExtensionSimpleRename(self):
        fn = self._makeInvalidRenameFunction(
          'mnt/%s/%s.%s' % (self.artist_name, self.track_name, 'xyz'))
        self.runTestFunction(fn)

    def _testInvalidDueToNonMatchingNameSimpleRename(self):
        fn = self._makeInvalidRenameFunction('mnt/boink')
        self.runTestFunction(fn)

################################################################################

class OggRenameTestCase(_BaseRenameTestCase):
    ext = 'ogg'

define_integration_test_classes(OggRenameTestCase)

class FlacRenameTestCase(_BaseRenameTestCase):
    ext = 'flac'

define_integration_test_classes(FlacRenameTestCase)

class Mp3RenameTestCase(_BaseRenameTestCase):
    ext = 'mp3'

define_integration_test_classes(Mp3RenameTestCase)

################################################################################

class _NonAsciiArtistBaseRenameTestCase(_BaseRenameTestCase):
    artist_name = u'•'

class NonAsciiArtistOggRenameTestCase(_NonAsciiArtistBaseRenameTestCase):
    ext = 'ogg'

define_integration_test_classes(NonAsciiArtistOggRenameTestCase)

class NonAsciiArtistFlacRenameTestCase(_NonAsciiArtistBaseRenameTestCase):
    ext = 'flac'

define_integration_test_classes(NonAsciiArtistFlacRenameTestCase)

class NonAsciiArtistMp3RenameTestCase(_NonAsciiArtistBaseRenameTestCase):
    ext = 'mp3'

define_integration_test_classes(NonAsciiArtistMp3RenameTestCase)

################################################################################

class MkdirRmdirTestCase(_BaseIntegrationTestCase):
    def get_opts(self):
        return 'format="/%g/%a/%t.%e"'

    def buildTree(self):
        data_file = '../../data/silence.flac'

        for num in (1, 2, 3, 4):
            self.sh.execute('cp ? ?', data_file, 'source/silence%u.flac' % num)

        for num, artist, title, genre in (
          (1, 'foo', 'bar', 'Rock'),
          (2, 'foo', 'baz', 'Rock'),
          (3, 'bink', 'bonk', 'Rap'),
          (4, 'fink', 'fonk', 'Country'),
        ):
            file = 'source/silence%u.flac' % num
            self.sh.execute(
              'pytags --set artist=? --set title=? --set genre=? ?',
              artist, title, genre, file,
            )

    def _testMkdirRmdir(self):
        def fn():
            dir = 'mnt/Honky Tonk'
            self.sh.execute('mkdir ?', dir)
            time.sleep(RACE_DELAY)
            self.assertDirectoryExists(dir)
            self.sh.execute('rmdir ?', dir)
            time.sleep(RACE_DELAY)
            self.assertDirectoryDoesNotExist(dir)
        self.runTestFunction(fn)

    def _testRenameEmptyDir(self):
        def fn():
            dir = 'mnt/Honky Tonk'
            dir2 = 'mnt/Soul'
            self.sh.execute('mkdir ?', dir)
            time.sleep(RACE_DELAY)
            self.assertDirectoryExists(dir)
            self.sh.execute('mv ? ?', dir, dir2)
            time.sleep(RACE_DELAY)
            self.assertDirectoryDoesNotExist(dir)
            self.sh.execute('ls ?', dir2)
            self.assertRaises(CommandFailed, self.sh.execute, 'ls ?/*', dir2)
            self.sh.execute('rmdir ?', dir2)
            time.sleep(RACE_DELAY)
            self.assertDirectoryDoesNotExist(dir2)
        self.runTestFunction(fn)

    def _testMoveFilesIntoNewDirectory(self):
        def fn():
            self.sh.execute('mkdir mnt/Rock/biz')
            time.sleep(RACE_DELAY)
            self.assertDirectoryExists('mnt/Rock/biz')

            self.sh.execute('mv mnt/Rock/foo/* mnt/Rock/biz/')
            time.sleep(RACE_DELAY)
            self.assertDirectoryDoesNotExist('mnt/Rock/foo')

            status, output = self.sh.execute(
              'pytags mnt/Rock/biz/* | grep artist=biz | wc -l')
            num_lines = int(output.strip())
            self.assertEqual(num_lines, 2)

        self.runTestFunction(fn)

define_integration_test_classes(MkdirRmdirTestCase)

################################################################################

class LogFileTestCase(_BaseIntegrationTestCase):
    def get_opts(self):
        return 'verbosity=debug'

    def _testLogFileExists(self):
        def fn():
            self.assertFileExists('mnt/.log')
        self.runTestFunction(fn)

    def _testLogFileReceivesMessages(self):
        def fn():
            from pytagsfs import __version__ as version
            self.assertFileContains('mnt/.log', 'pytagsfs version %s' % version)
        self.runTestFunction(fn)

define_integration_test_classes(LogFileTestCase)

class LogSizeTestCase(_BaseIntegrationTestCase):
    logsize = 8

    def get_opts(self):
        return 'verbosity=debug,logsize=%u' % self.logsize

    def _testLogSize(self):
        def fn():
            f = open('mnt/.log', 'r')
            length = len(f.read())
            try:
                assert length == self.logsize, (
                  'Expected length to be %u, got %u instead.' % (
                    self.logsize, length))
            finally:
                f.close()
        self.runTestFunction(fn)

define_integration_test_classes(LogSizeTestCase)

if __name__ == '__main__':
    main()
