# coding: utf-8

# Copyright (c) 2007-2008 Forest Bond.
# This file is part of the pytagsfs software package.
#
# pytagsfs is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

from unittest import TestCase, main

from pytagsfs.pathparser import PathParser
from pytagsfs.sourcetree import SourceTree
from pytagsfs.sourcetreemon import SourceTreeMonitor
from pytagsfs.sourcetreerep import SourceTreeRepresentation
from pytagsfs.metaparser import DelegateMultiMetaParser, FilenameMetaParser, \
  ParentDirMetaParser, ExtensionMetaParser
from pytagsfs.pathstore.pytypes import PyTypesPathStore
from pytagsfs.exceptions import PathNotFound

import os

class MockSourceTree(SourceTree):
    tree_entries = {
      '/': (['a', u'•'], ['c']),
      '/a': ([], ['b']),
      u'/•': ([], ['x']),
    }

    def walk(self, path = None):
        if path is None:
            path = self.root
        path = path.rstrip('/')

        rel_path = self.get_relative_path(path)
        if rel_path == '':
            rel_path = '/'

        sub_paths = [ ]

        dirpath = path
        dirnames = self.tree_entries[rel_path][0]
        filenames = self.tree_entries[rel_path][1]

        yield (dirpath, dirnames, filenames)

        for dirname in dirnames:
            sub_paths.append(os.path.join(dirpath, dirname))

        for sub_path in sub_paths:
            for dirpath, dirnames, filenames in self.walk(sub_path):
                yield (dirpath, dirnames, filenames)

    def isreadable(self, path):
        if path in (
          u'%s/c' % self.root,
          u'%s/a/b' % self.root,
          u'%s/•/x' % self.root,
        ):
            return True
        return False

class MockSourceTreeMonitor(SourceTreeMonitor):
    pass

class SourceTreeRepresentationTestCase(TestCase):
    def testSanity(self):
        root = '/foo/bar'
        format_string = '/%p/%f'
        meta_parser = DelegateMultiMetaParser((
          FilenameMetaParser(),
          ParentDirMetaParser(),
          ExtensionMetaParser(),
        ))
        rep = SourceTreeRepresentation(
          path_parser = PathParser(meta_parser, format_string),
          path_store = PyTypesPathStore(),
          source_tree = MockSourceTree(root),
          monitor = MockSourceTreeMonitor(),
        )

        rep.start()

        self.assertEqual(rep.get_real_path(u'/•/x'), os.path.join(root, u'•/x'))
        self.assertRaises(PathNotFound, rep.get_real_path, '/a/x')

        self.assertEqual(rep.get_entries('/'), ['a', u'•', 'bar'])
        self.assertEqual(rep.get_entries(u'/•'), ['x'])

        rep.stop()

import common

common.defineDocTestCases(__name__, 'pytagsfs.sourcetree')

if __name__ == '__main__':
    main()
